@@ -1,6 +1,4 @@
 /*
- * $Id: init.c,v 2.46 1996/10/15 20:16:35 hzoli Exp $
- *
  * init.c - main loop and initialization routines
  *
  * This file is part of zsh, the Z shell.
@@ -74,12 +72,12 @@ main(int argc, char **argv)
 
     for (;;) {
 	do
-	    loop(1);
+	    loop(1,0);
 	while (tok != ENDINPUT);
 	if (!(isset(IGNOREEOF) && interact)) {
 #if 0
 	    if (interact)
-		fputs(islogin ? "logout\n" : "exit\n", stderr);
+		fputs(islogin ? "logout\n" : "exit\n", shout);
 #endif
 	    zexit(lastval, 0);
 	    continue;
@@ -98,7 +96,7 @@ main(int argc, char **argv)
 
 /**/
 void
-loop(int toplevel)
+loop(int toplevel, int justonce)
 {
     List list;
 #ifdef DEBUG
@@ -119,13 +117,27 @@ loop(int toplevel)
 	lexinit();              /* initialize lexical state */
 	if (!(list = parse_event())) {	/* if we couldn't parse a list */
 	    hend();
-	    if (tok == ENDINPUT && !errflag)
+	    if ((tok == ENDINPUT && !errflag) || justonce)
 		break;
 	    continue;
 	}
 	if (hend()) {
 	    int toksav = tok;
+	    List prelist;
 
+	    if (toplevel && (prelist = getshfunc("preexec")) != &dummy_list) {
+		Histent he = gethistent(curhist);
+		LinkList args;
+		PERMALLOC {
+		    args = newlinklist();
+		    addlinknode(args, "preexec");
+		    if (he && he->text)
+			addlinknode(args, he->text);
+		} LASTALLOC;
+		doshfunc(prelist, args, 0, 1);
+		freelinklist(args, (FreeFunc) NULL);
+		errflag = 0;
+	    }
 	    if (stopmsg)	/* unset 'you have stopped jobs' flag */
 		stopmsg--;
 	    execlist(list, 0, 0);
@@ -151,6 +163,8 @@ loop(int toplevel)
 		dotrap(SIGEXIT);
 	    exit(lastval);
 	}
+	if (justonce)
+	    break;
     }
     popheap();
 }
@@ -208,6 +222,7 @@ parseargs(char **argv)
 
     /* loop through command line options (begins with "-" or "+") */
     while (*argv && (**argv == '-' || **argv == '+')) {
+	char *args = *argv;
 	action = (**argv == '-');
 	if(!argv[0][1])
 	    *argv = "--";
@@ -221,14 +236,9 @@ parseargs(char **argv)
 	    }
 
 	    if (**argv == 'c') {         /* -c command */
-		if (!*++argv) {
-		    zerr("string expected after -c", NULL, 0);
-		    exit(1);
-		}
-		cmd = *argv++;
+		cmd = *argv;
 		opts[INTERACTIVE] &= 1;
 		opts[SHINSTDIN] = 0;
-		goto doneoptions;
 	    } else if (**argv == 'o') {
 		if (!*++*argv)
 		    argv++;
@@ -240,6 +250,14 @@ parseargs(char **argv)
 		    zerr("no such option: %s", *argv, 0);
 		else
 		    dosetopt(optno, action, 1);
+              break;
+	    } else if (isspace(**argv)) {
+		/* zsh's typtab not yet set, have to use ctype */
+		while (*++*argv)
+		    if (!isspace(**argv)) {
+			zerr("bad option string: `%s'", args, 0);
+			exit(1);
+		    }
 		break;
 	    } else {
 	    	if (!(optno = optlookupc(**argv))) {
@@ -253,6 +271,13 @@ parseargs(char **argv)
     }
     doneoptions:
     paramlist = newlinklist();
+    if (cmd) {
+	if (!*argv) {
+	    zerr("string expected after -%s", cmd, 0);
+	    exit(1);
+	}
+	cmd = *argv++;
+    }
     if (*argv) {
 	if (unset(SHINSTDIN)) {
 	    argzero = *argv;
@@ -322,15 +347,46 @@ init_io(void)
     if (isatty(0)) {
 	zsfree(ttystrname);
 	if ((ttystrname = ztrdup(ttyname(0))))
-	    SHTTY = movefd(open(ttystrname, O_RDWR));
+	    SHTTY = movefd(open(ttystrname, O_RDWR | O_NOCTTY));
+	/*
+	 * xterm, rxvt and probably all terminal emulators except
+	 * dtterm on Solaris 2.6 & 7 have a bug. Applications are
+	 * unable to open /dev/tty or /dev/pts/<terminal number here>
+	 * because something in Sun's STREAMS modules doesn't like
+	 * it. The open() call fails with EBUSY which is not even
+	 * listed as a possibility in the open(2) man page.  So we'll
+	 * try to outsmart The Company.  -- <dave@srce.hr>
+	 *
+	 * Presumably there's no harm trying this on any OS, given that
+	 * isatty(0) worked but opening the tty didn't.  Possibly we won't
+	 * get the tty read/write, but it's the best we can do -- pws
+	 *
+	 * Try both stdin and stdout before trying /dev/tty. -- Bart
+	 */
+#if defined(HAVE_FCNTL_H) && defined(F_GETFL)
+#define rdwrtty(fd)	((fcntl(fd, F_GETFL, 0) & O_RDWR) == O_RDWR)
+#else
+#define rdwrtty(fd)	1
+#endif
+	if (SHTTY == -1 && rdwrtty(0)) {
+	    SHTTY = movefd(dup(0));
+	}
     }
-    if (SHTTY == -1 && (SHTTY = movefd(open("/dev/tty", O_RDWR))) != -1) {
+    if (SHTTY == -1 && isatty(1) && rdwrtty(1) &&
+	(SHTTY = movefd(dup(1))) != -1) {
 	zsfree(ttystrname);
-	ttystrname = ztrdup("/dev/tty");
+	ttystrname = ztrdup(ttyname(1));
+    }
+    if (SHTTY == -1 &&
+	(SHTTY = movefd(open("/dev/tty", O_RDWR | O_NOCTTY))) != -1) {
+	zsfree(ttystrname);
+	ttystrname = ztrdup(ttyname(SHTTY));
     }
     if (SHTTY == -1) {
 	zsfree(ttystrname);
 	ttystrname = ztrdup("");
+    } else if (!ttystrname) {
+	ttystrname = ztrdup("/dev/tty");
     }
 
     /* We will only use zle if shell is interactive, *
@@ -492,6 +548,7 @@ setupvals(void)
     int i;
 #endif
 
+    lineno = 1;
     noeval = 0;
     curhist = 0;
     histsiz = DEFAULT_HISTSIZE;
@@ -549,8 +606,8 @@ setupvals(void)
     zoptind = 1;
     schedcmds = NULL;
 
-    ppid  = (long) getppid();
-    mypid = (long) getpid();
+    ppid  = (zlong) getppid();
+    mypid = (zlong) getpid();
     term  = ztrdup("");
 
     /* The following variable assignments cause zsh to behave more *
@@ -714,7 +771,7 @@ run_init_scripts(void)
 	    char *s = getsparam("ENV");
 	    if (islogin)
 		sourcehome(".profile");
-	    noerrs = 1;
+	    noerrs = 2;
 	    if (s && !parsestr(s)) {
 		singsub(&s);
 		noerrs = 0;
@@ -804,13 +861,13 @@ source(char *s)
     SHIN = tempfd;
     bshin = fdopen(SHIN, "r");
     subsh  = 0;
-    lineno = 0;
+    lineno = 1;
     loops  = 0;
     dosetopt(SHINSTDIN, 0, 1);
     scriptname = s;
 
     sourcelevel++;
-    loop(0);			/* loop through the file to be sourced        */
+    loop(0, 0);			/* loop through the file to be sourced        */
     sourcelevel--;
     fclose(bshin);
     fdtable[SHIN] = 0;
