@@ -1,6 +1,4 @@
 /*
- * $Id: loop.c,v 2.4 1996/10/15 20:16:35 hzoli Exp $
- *
  * loop.c - loop execution
  *
  * This file is part of zsh, the Z shell.
@@ -40,8 +38,6 @@ execfor(Cmd cmd)
     char *str;
     LinkList args;
 
-    loops++;
-    lastval = 0;
     node = cmd->u.forcmd;
     args = cmd->args;
     if (!node->inflag) {
@@ -51,6 +47,8 @@ execfor(Cmd cmd)
 	for (x = pparams; *x; x++)
 	    addlinknode(args, ztrdup(*x));
     }
+    lastval = 0;
+    loops++;
     pushheap();
     while ((str = (char *)ugetnode(args))) {
 	setsparam(node->name, ztrdup(str));
@@ -104,9 +102,15 @@ execselect(Cmd cmd)
 	do {
 	    selectlist(args);
 	    if (empty(bufstack)) {
-	    	if (interact && SHTTY != -1 && isset(USEZLE))
+	    	if (interact && SHTTY != -1 && isset(USEZLE)) {
+		    int oef = errflag;
+
+		    isfirstln = 1;
 		    str = (char *)zleread(prompt3, NULL);
-	    	else {
+		    if (errflag)
+			str = NULL;
+		    errflag = oef;
+	    	} else {
 		    int pptlen;
 		    str = putprompt(prompt3, &pptlen, NULL, 1);
 		    fwrite(str, pptlen, 1, stderr);
@@ -168,11 +172,10 @@ execwhile(Cmd cmd)
 
     olderrexit = noerrexit;
     node = cmd->u.whilecmd;
-    lastval = 0;
+    oldval = 0;
     pushheap();
     loops++;
     for (;;) {
-	oldval = lastval;
 	list = (List) dupstruct(node->cont);
 	noerrexit = 1;
 	execlist(list, 1, 0);
@@ -196,6 +199,7 @@ execwhile(Cmd cmd)
 	    lastval = 1;
 	    break;
 	}
+	oldval = lastval;
     }
     popheap();
     loops--;
@@ -275,24 +279,17 @@ execcase(Cmd cmd)
 {
     struct casecmd *node;
     char *word;
-    LinkList args;
     List *l;
     char **p;
 
     node = cmd->u.casecmd;
-    args = cmd->args;
     l = node->lists;
     p = node->pats;
 
+    word = *p++;
+    singsub(&word);
+    untokenize(word);
     lastval = 0;
-    if (firstnode(args) && nextnode(firstnode(args))) {
-	zerr("too many arguments to case", NULL, 0);
-	return 1;
-    }
-    if (empty(args))
-	word = dupstring("");
-    else
-	word = (char *)peekfirst(args);
 
     if (node) {
 	while (*p) {
