@@ -1,11 +1,9 @@
 /*
- * $Id: utils.c,v 2.61 1996/10/15 20:16:35 hzoli Exp $
- *
  * utils.c - miscellaneous utilities
  *
  * This file is part of zsh, the Z shell.
  *
- * Copyright (c) 1992-1996 Paul Falstad
+ * Copyright (c) 1992-1997 Paul Falstad
  * All rights reserved.
  *
  * Permission is hereby granted, without written agreement and without
@@ -32,53 +30,70 @@
 #include "zsh.h"
 
 /* Print an error */
-
-/**/
-void
-zwarnnam(char *cmd, char *fmt, char *str, int num)
-{
-    int waserr;
-
-    waserr = errflag;
-    zerrnam(cmd, fmt, str, num);
-    errflag = waserr;
-}
-
+ 
 /**/
 void
 zerr(char *fmt, char *str, int num)
 {
-    if (errflag || noerrs)
+    if (errflag || noerrs) {
+	if (noerrs < 2)
+	    errflag = 1;
 	return;
+    }
+    zwarn(fmt, str, num);
     errflag = 1;
-    trashzle();
-    /*
-     * scriptname is set when sourcing scripts, so that we get the
-     * correct name instead of the generic name of whatever
-     * program/script is running.
-     */
-    nicezputs(isset(SHINSTDIN) ? "zsh" :
-	      scriptname ? scriptname : argzero, stderr);
-    fputs(": ", stderr);
-    zerrnam(NULL, fmt, str, num);
 }
 
 /**/
 void
 zerrnam(char *cmd, char *fmt, char *str, int num)
 {
-    if (cmd) {
-	if (errflag || noerrs)
-	    return;
-	errflag = 1;
-	trashzle();
-	if(unset(SHINSTDIN)) {
-	    nicezputs(scriptname ? scriptname : argzero, stderr);
-	    fputs(": ", stderr);
-	}
-	nicezputs(cmd, stderr);
+    if (errflag || noerrs)
+	return;
+
+    zwarnnam(cmd, fmt, str, num);
+    errflag = 1;
+}
+
+/**/
+void
+zwarn(char *fmt, char *str, int num)
+{
+    if (errflag || noerrs)
+	return;
+    trashzle();
+    /*
+     * scriptname is set when sourcing scripts, so that we get the
+     * correct name instead of the generic name of whatever
+     * program/script is running.  It's also set in shell functions,
+     * so test locallevel, too.
+     */
+    nicezputs((isset(SHINSTDIN) && !locallevel) ? "zsh" :
+	      scriptname ? scriptname : argzero, stderr);
+    fputs(": ", stderr);
+    zerrmsg(fmt, str, num);
+}
+
+/**/
+void
+zwarnnam(char *cmd, char *fmt, char *str, int num)
+{
+    if (errflag || noerrs)
+	return;
+    trashzle();
+    if (unset(SHINSTDIN) || locallevel) {
+	nicezputs(scriptname ? scriptname : argzero, stderr);
 	fputs(": ", stderr);
     }
+    nicezputs(cmd, stderr);
+    fputs(": ", stderr);
+    zerrmsg(fmt, str, num);
+}
+
+/**/
+void
+zerrmsg(char *fmt, char *str, int num)
+{
     while (*fmt)
 	if (*fmt == '%') {
 	    fmt++;
@@ -87,12 +102,12 @@ zerrnam(char *cmd, char *fmt, char *str, int num)
 		nicezputs(str, stderr);
 		break;
 	    case 'l': {
-		char sav;
+		char *s;
 		num = metalen(str, num);
-		sav = str[num];
-		str[num] = '\0';
-		nicezputs(str, stderr);
-		str[num] = sav;
+		s = halloc(num + 1);
+		memcpy(s, str, num);
+		s[num] = '\0';
+		nicezputs(s, stderr);
 		break;
 	    }
 	    case 'd':
@@ -126,8 +141,8 @@ zerrnam(char *cmd, char *fmt, char *str, int num)
 	    putc(*fmt == Meta ? *++fmt ^ 32 : *fmt, stderr);
 	    fmt++;
 	}
-    if (unset(SHINSTDIN) && lineno)
-	fprintf(stderr, " [%ld]\n", lineno);
+    if ((unset(SHINSTDIN) || locallevel) && lineno)
+	fprintf(stderr, " [%ld]\n", (long)lineno);
     else
 	putc('\n', stderr);
     fflush(stderr);
@@ -174,6 +189,8 @@ nicechar(int c)
     if (isprint(c))
 	goto done;
     if (c & 0x80) {
+	if (isset(PRINTEIGHTBIT))
+	    goto done;
 	*s++ = '\\';
 	*s++ = 'M';
 	*s++ = '-';
@@ -200,10 +217,9 @@ nicechar(int c)
     return buf;
 }
 
-#if 0
 /* Output a string's visible representation. */
 
-/**/
+#if 0 /**/
 void
 nicefputs(char *s, FILE *f)
 {
@@ -296,7 +312,7 @@ xsymlinks(char *s, int flag)
 {
     char **pp, **opp;
     char xbuf2[PATH_MAX*2], xbuf3[PATH_MAX*2];
-    int t0;
+    int t0, ret = 0;
 
     opp = pp = slashsplit(s);
     for (; *pp; pp++) {
@@ -331,15 +347,14 @@ xsymlinks(char *s, int flag)
 	    metafy(xbuf3, t0, META_NOALLOC);
 	    if (*xbuf3 == '/') {
 		strcpy(xbuf, "");
-		if (xsymlinks(xbuf3 + 1, flag))
-		    return 1;
-	    } else if (xsymlinks(xbuf3, flag))
-		return 1;
+		ret = xsymlinks(xbuf3 + 1, flag);
+	    } else
+		ret = xsymlinks(xbuf3, flag);
 	    zsfree(*pp);
 	}
     }
     free(opp);
-    return 0;
+    return ret;
 }
 
 /* expand symlinks in s, and remove other weird things */
@@ -353,7 +368,7 @@ xsymlink(char *s)
     if (*s != '/')
 	return NULL;
     *xbuf = '\0';
-    if (xsymlinks(s + 1, 1))
+    if (!xsymlinks(s + 1, 1))
 	return ztrdup(s);
     if (!*xbuf)
 	return ztrdup("/");
@@ -381,10 +396,16 @@ fprintdir(char *s, FILE *f)
  * and uid to try to avoid requerying the password files *
  * or NIS/NIS+ database.                                 */
 
+/**/
+uid_t cached_uid;
+/**/
+char *cached_username;
+
 /**/
 char *
 get_username(void)
 {
+#ifdef HAVE_GETPWUID
     struct passwd *pswd;
     uid_t current_uid;
  
@@ -397,6 +418,9 @@ get_username(void)
 	else
 	    cached_username = ztrdup("");
     }
+#else /* !HAVE_GETPWUID */
+    cached_uid = getuid();
+#endif /* !HAVE_GETPWUID */
     return cached_username;
 }
 
@@ -507,7 +531,6 @@ getnameddir(char *name)
 {
     Param pm;
     char *str;
-    struct passwd *pw;
     Nameddir nd;
 
     /* Check if it is already in the named directory table */
@@ -524,14 +547,19 @@ getnameddir(char *name)
 	return str;
     }
 
-    /* Retrieve an entry from the password table/database for this user. */
-    if ((pw = getpwnam(name))) {
-	char *dir = xsymlink(pw->pw_dir);
-	adduserdir(name, dir, ND_USERNAME, 1);
-	str = dupstring(dir);
-	zsfree(dir);
-	return str;
+#ifdef HAVE_GETPWNAM
+    {
+	/* Retrieve an entry from the password table/database for this user. */
+	struct passwd *pw;
+	if ((pw = getpwnam(name))) {
+	    char *dir = xsymlink(pw->pw_dir);
+	    adduserdir(name, dir, ND_USERNAME, 1);
+	    str = dupstring(dir);
+	    zsfree(dir);
+	    return str;
+	}
     }
+#endif /* HAVE_GETPWNAM */
 
     /* There are no more possible sources of directory names, so give up. */
     return NULL;
@@ -688,10 +716,10 @@ checkmailpath(char **s)
 	    }
 	} else {
 	    if (st.st_size && st.st_atime <= st.st_mtime &&
-		st.st_mtime > lastmailcheck)
+		st.st_mtime > lastmailcheck) {
 		if (!u) {
-		    fprintf(stderr, "You have new mail.\n");
-		    fflush(stderr);
+		    fprintf(shout, "You have new mail.\n");
+		    fflush(shout);
 		} else {
 		    char *usav = underscore;
 
@@ -700,17 +728,18 @@ checkmailpath(char **s)
 			u = dupstring(u);
 			if (! parsestr(u)) {
 			    singsub(&u);
-			    zputs(u, stderr);
-			    fputc('\n', stderr);
-			    fflush(stderr);
+			    zputs(u, shout);
+			    fputc('\n', shout);
+			    fflush(shout);
 			}
 			underscore = usav;
 		    } LASTALLOC;
 		}
+	    }
 	    if (isset(MAILWARNING) && st.st_atime > st.st_mtime &&
 		st.st_atime > lastmailcheck && st.st_size) {
-		fprintf(stderr, "The mail in %s has been read.\n", unmeta(*s));
-		fflush(stderr);
+		fprintf(shout, "The mail in %s has been read.\n", unmeta(*s));
+		fflush(shout);
 	    }
 	}
 	*v = c;
@@ -828,61 +857,131 @@ settyinfo(struct ttyinfo *ti)
 extern winchanged;
 #endif
 
-/* check the size of the window and adjust if necessary. *
- * The value of from:					 *
- *   0: called from update_job or setupvals		 *
- *   1: called from the SIGWINCH handler		 *
- *   2: the user have just changed LINES manually	 *
- *   3: the user have just changed COLUMNS manually      */
-
-/**/
-void
-adjustwinsize(int from)
+static int
+adjustlines(int signalled)
 {
-    int oldcols = columns, oldrows = lines;
+    int oldlines = lines;
 
 #ifdef TIOCGWINSZ
-    static int userlines, usercols;
-
-    if (SHTTY == -1)
-	return;
-
-    if (from == 2)
-	userlines = lines > 0;
-    if (from == 3)
-	usercols = columns > 0;
-
-    if (!ioctl(SHTTY, TIOCGWINSZ, (char *)&shttyinfo.winsize)) {
-	if (!userlines)
-	    lines = shttyinfo.winsize.ws_row;
-	if (!usercols)
-	    columns = shttyinfo.winsize.ws_col;
-    }
-#endif   /* TIOCGWINSZ */
-
-    if (lines <= 0)
+    if (signalled || lines <= 0)
+	lines = shttyinfo.winsize.ws_row;
+    else
+	shttyinfo.winsize.ws_row = lines;
+#endif /* TIOCGWINSZ */
+    if (lines <= 0) {
+	DPUTS(signalled, "BUG: Impossible TIOCGWINSZ rows");
 	lines = tclines > 0 ? tclines : 24;
-    if (columns <= 0)
-	columns = tccolumns > 0 ? tccolumns : 80;
+    }
+
     if (lines > 2)
 	termflags &= ~TERM_SHORT;
     else
 	termflags |= TERM_SHORT;
+
+    return (lines != oldlines);
+}
+
+static int
+adjustcolumns(int signalled)
+{
+    int oldcolumns = columns;
+
+#ifdef TIOCGWINSZ
+    if (signalled || columns <= 0)
+	columns = shttyinfo.winsize.ws_col;
+    else
+	shttyinfo.winsize.ws_col = columns;
+#endif /* TIOCGWINSZ */
+    if (columns <= 0) {
+	DPUTS(signalled, "BUG: Impossible TIOCGWINSZ cols");
+	columns = tccolumns > 0 ? tccolumns : 80;
+    }
+
     if (columns > 2)
 	termflags &= ~TERM_NARROW;
     else
 	termflags |= TERM_NARROW;
 
+    return (columns != oldcolumns);
+}
+
+/* check the size of the window and adjust if necessary. *
+ * The value of from:					 *
+ *   0: called from update_job or setupvals		 *
+ *   1: called from the SIGWINCH handler		 *
+ *   2: called from the LINES parameter callback	 *
+ *   3: called from the COLUMNS parameter callback	 */
+
+/**/
+void
+adjustwinsize(int from)
+{
+    static int getwinsz = 1;
+    int ttyrows = shttyinfo.winsize.ws_row;
+    int ttycols = shttyinfo.winsize.ws_col;
+    int resetzle = 0;
+
+    if (getwinsz || from == 1) {
 #ifdef TIOCGWINSZ
-    if (from >= 2) {
-	shttyinfo.winsize.ws_row = lines;
-	shttyinfo.winsize.ws_col = columns;
+	if (SHTTY == -1)
+	    return;
+	if (ioctl(SHTTY, TIOCGWINSZ, (char *)&shttyinfo.winsize) == 0) {
+	    resetzle = (ttyrows != shttyinfo.winsize.ws_row ||
+			ttycols != shttyinfo.winsize.ws_col);
+	    if (from == 0 && resetzle && ttyrows && ttycols)
+		from = 1; /* Signal missed while a job owned the tty? */
+	    ttyrows = shttyinfo.winsize.ws_row;
+	    ttycols = shttyinfo.winsize.ws_col;
+	} else {
+	    /* Set to unknown on failure */
+	    shttyinfo.winsize.ws_row = 0;
+	    shttyinfo.winsize.ws_col = 0;
+	    resetzle = 1;
+	}
+#else
+	resetzle = from == 1;
+#endif /* TIOCGWINSZ */
+    } /* else
+	 return; */
+
+    switch (from) {
+    case 0:
+    case 1:
+	getwinsz = 0;
+	/* Calling setiparam() here calls this function recursively, but  *
+	 * because we've already called adjustlines() and adjustcolumns() *
+	 * here, recursive calls are no-ops unless a signal intervenes.   *
+	 * The commented "else return;" above might be a safe shortcut,   *
+	 * but I'm concerned about what happens on race conditions; e.g., *
+	 * suppose the user resizes his xterm during `eval $(resize)'?    */
+	if (adjustlines(from) && zgetenv("LINES"))
+	    setiparam("LINES", lines);
+	if (adjustcolumns(from) && zgetenv("COLUMNS"))
+	    setiparam("COLUMNS", columns);
+	getwinsz = 1;
+	break;
+    case 2:
+	resetzle = adjustlines(0);
+	break;
+    case 3:
+	resetzle = adjustcolumns(0);
+	break;
+    }
+
+#ifdef TIOCGWINSZ
+    if (interact && from >= 2 &&
+	(shttyinfo.winsize.ws_row != ttyrows ||
+	 shttyinfo.winsize.ws_col != ttycols)) {
+	/* shttyinfo.winsize is already set up correctly */
 	ioctl(SHTTY, TIOCSWINSZ, (char *)&shttyinfo.winsize);
     }
-#endif
+#endif /* TIOCGWINSZ */
 
-    if (zleactive && (from >= 2 || oldcols != columns || oldrows != lines)) {
-	resetneeded = winchanged = 1;
+    if (zleactive && resetzle) {
+#ifdef TIOCGWINSZ
+	winchanged =
+#endif /* TIOCGWINSZ */
+	    resetneeded = 1;
 	refresh();
     }
 }
@@ -1071,15 +1170,15 @@ skipparens(char inpar, char outpar, char **s)
    return level;
 }
 
-/* Convert string to long.  This function (without the z) *
- * is contained in the ANSI standard C library, but a lot *
- * of them seem to be broken.                             */
+/* Convert string to zlong.  This function (without the z) *
+ * is contained in the ANSI standard C library, but a lot  *
+ * of them seem to be broken.                              */
 
 /**/
-long
+zlong
 zstrtol(const char *s, char **t, int base)
 {
-    long ret = 0;
+    zlong ret = 0;
     int neg;
 
     while (inblank(*s))
@@ -1090,14 +1189,15 @@ zstrtol(const char *s, char **t, int base)
     else if (*s == '+')
 	s++;
 
-    if (!base)
+    if (!base) {
 	if (*s != '0')
 	    base = 10;
 	else if (*++s == 'x' || *s == 'X')
 	    base = 16, s++;
 	else
 	    base = 8;
- 
+    }
+
     if (base <= 10)
 	for (; *s >= '0' && *s < ('0' + base); s++)
 	    ret = ret * base + *s - '0';
@@ -1120,14 +1220,15 @@ zstrtorlimit(const char *s, char **t, int base)
 {
     rlim_t ret = 0;
  
-    if (!base)
+    if (!base) {
 	if (*s != '0')
 	    base = 10;
 	else if (*++s == 'x' || *s == 'X')
 	    base = 16, s++;
 	else
 	    base = 8;
- 
+    }
+
     if (base <= 10)
 	for (; *s >= '0' && *s < ('0' + base); s++)
 	    ret = ret * base + *s - '0';
@@ -1162,25 +1263,54 @@ checkrmall(char *s)
 int
 setblock_stdin(void)
 {
-#ifdef O_NONBLOCK
+#ifdef O_NDELAY
+# ifdef O_NONBLOCK
+#  define NONBLOCK (O_NDELAY|O_NONBLOCK)
+# else /* !O_NONBLOCK */
+#  define NONBLOCK O_NDELAY
+# endif /* !O_NONBLOCK */
+#else /* !O_NDELAY */
+# ifdef O_NONBLOCK
+#  define NONBLOCK O_NONBLOCK
+# else /* !O_NONBLOCK */
+#  define NONBLOCK 0
+# endif /* !O_NONBLOCK */
+#endif /* !O_NDELAY */
+
+#if NONBLOCK
     struct stat st;
     long mode;
 
     if (!fstat(0, &st) && !S_ISREG(st.st_mode)) {
-	mode = fcntl(0, F_GETFL);
-	if (mode != -1 && (mode & O_NONBLOCK) &&
-	    !fcntl(0, F_SETFL, mode & ~O_NONBLOCK))
+	mode = fcntl(0, F_GETFL, 0);
+	if (mode != -1 && (mode & NONBLOCK) &&
+	    !fcntl(0, F_SETFL, mode & ~NONBLOCK))
 	    return 1;
     }
-#endif
+#endif /* NONBLOCK */
     return 0;
+
+#undef NONBLOCK
+}
+
+/**/
+int
+read1char(void)
+{
+    char c;
+
+    while (read(SHTTY, &c, 1) != 1) {
+	if (errno != EINTR || errflag)
+	    return -1;
+    }
+    return STOUC(c);
 }
 
 /**/
 int
 getquery(char *valid_chars)
 {
-    char c, d;
+    int c, d;
     int isem = !strcmp(term, "emacs");
 
 #ifdef FIONREAD
@@ -1200,7 +1330,7 @@ getquery(char *valid_chars)
 	return 'n';
     }
 #endif
-    while (read(SHTTY, &c, 1) == 1) {
+    while ((c = read1char()) >= 0) {
 	if (c == 'Y' || c == '\t')
 	    c = 'y';
 	else if (c == 'N')
@@ -1222,13 +1352,13 @@ getquery(char *valid_chars)
     }
     if (isem) {
 	if (c != '\n')
-	    while (read(SHTTY, &d, 1) == 1 && d != '\n');
+	    while ((d = read1char()) >= 0 && d != '\n');
     } else {
 	settyinfo(&shttyinfo);
 	if (c != '\n' && !valid_chars)
 	    write(2, "\n", 1);
     }
-    return (int)c;
+    return c;
 }
 
 static int d;
@@ -1313,7 +1443,7 @@ spckword(char **s, int hist, int cmd, int ask)
 		return;
 	    guess = dupstring(guess);
 	    ne = noerrs;
-	    noerrs = 1;
+	    noerrs = 2;
 	    singsub(&guess);
 	    noerrs = ne;
 	    if (!guess)
@@ -1437,7 +1567,7 @@ ztrftime(char *buf, int bufsize, char *fmt, struct tm *tm)
 		*buf++ = '0' + tm->tm_sec % 10;
 		break;
 	    case 'y':
-		*buf++ = '0' + tm->tm_year / 10;
+		*buf++ = '0' + (tm->tm_year / 10) % 10;
 		*buf++ = '0' + tm->tm_year % 10;
 		break;
 #ifndef HAVE_STRFTIME
@@ -1606,10 +1736,11 @@ findsep(char **s, char *sep)
 	if (!*t)
 	    return i;
 	if (*(*s)++ == Meta) {
-	    (*s)++;
 #ifdef DEBUG
-	    if (! **s)
+	    if (! *(*s)++)
 		fprintf(stderr, "BUG: unexpected end of string in findsep()\n");
+#else
+	    (*s)++;
 #endif
 	}
     }
@@ -1692,13 +1823,12 @@ char *
 sepjoin(char **s, char *sep)
 {
     char *r, *p, **t;
-    int l, sl, elide = 0;
+    int l, sl;
     char sepbuf[3];
 
     if (!*s)
 	return "";
     if (!sep) {
-	elide = 1;
 	sep = sepbuf;
 	sepbuf[0] = *ifs;
 	sepbuf[1] = *ifs == Meta ? ifs[1] ^ 32 : '\0';
@@ -2134,22 +2264,22 @@ dupnode(int type, void *a, int argnum)
 	return (useheap) ? ((void *) dupstring(a)) :
 	    ((void *) ztrdup(a));
     case NT_LIST | NT_NODE:
-	if (type & NT_HEAP)
+	if (type & NT_HEAP) {
 	    if (useheap)
 		return (void *) duplist(a, (VFunc) dupstruct2);
 	    else
 		return (void *) list2arr(a, (VFunc) dupstruct2);
-	else if (useheap)
+	} else if (useheap)
 	    return (void *) arr2list(a, (VFunc) dupstruct2);
 	else
 	    return (void *) duparray(a, (VFunc) dupstruct2);
     case NT_LIST | NT_STR:
-	if (type & NT_HEAP)
+	if (type & NT_HEAP) {
 	    if (useheap)
 		return (void *) duplist(a, (VFunc) dupstring);
 	    else
 		return (void *) list2arr(a, (VFunc) ztrdup);
-	else if (useheap)
+	} else if (useheap)
 	    return (void *) arr2list(a, (VFunc) dupstring);
 	else
 	    return (void *) duparray(a, (VFunc) ztrdup);
@@ -2593,11 +2723,12 @@ inittyptab(void)
     for (t0 = (int)STOUC(Pound); t0 <= (int)STOUC(Nularg); t0++)
 	typtab[t0] |= ITOK | IMETA;
     for (s = ifs ? ifs : DEFAULT_IFS; *s; s++) {
-	if (inblank(*s))
+	if (inblank(*s)) {
 	    if (s[1] == *s)
 		s++;
 	    else
 		typtab[STOUC(*s)] |= IWSEP;
+	}
 	typtab[STOUC(*s == Meta ? *++s ^ 32 : *s)] |= ISEP;
     }
     for (s = wordchars ? wordchars : DEFAULT_WORDCHARS; *s; s++)
@@ -2789,7 +2920,7 @@ attachtty(pid_t pgrp)
 # endif
 #endif
 	{
-	    if (pgrp != mypgrp && kill(pgrp, 0) == -1)
+	    if (pgrp != mypgrp && kill(-pgrp, 0) == -1)
 		attachtty(mypgrp);
 	    else {
 		if (errno != ENOTTY)
@@ -3209,11 +3340,12 @@ nicezputs(char const *s, FILE *stream)
     int c;
 
     while ((c = *s++)) {
-	if (itok(c))
+	if (itok(c)) {
 	    if (c <= Comma)
 		c = ztokens[c - Pound];
 	    else 
 		continue;
+	}
 	if (c == Meta)
 	    c = *s++ ^ 32;
 	if(fputs(nicechar(c), stream) < 0)
@@ -3232,11 +3364,12 @@ niceztrlen(char const *s)
     int c;
 
     while ((c = *s++)) {
-	if (itok(c))
+	if (itok(c)) {
 	    if (c <= Comma)
 		c = ztokens[c - Pound];
 	    else 
 		continue;
+	}
 	if (c == Meta)
 	    c = *s++ ^ 32;
 	l += strlen(nicechar(STOUC(c)));
@@ -3401,6 +3534,8 @@ dosetopt(int optno, int value, int force)
     }
     if(!force && (optno == INTERACTIVE || optno == SHINSTDIN ||
 	    optno == SINGLECOMMAND)) {
+	if (opts[optno] == value)
+	    return 0;
 	/* it is not permitted to change the value of these options */
 	return -1;
     } else if(!force && optno == USEZLE && value) {
