@@ -1,6 +1,4 @@
 /*
- * $Id: zsh.h,v 2.49 1996/10/15 20:16:35 hzoli Exp $
- *
  * zsh.h - standard header file
  *
  * This file is part of zsh, the Z shell.
@@ -33,6 +31,31 @@
  * using -I. -I$srcdir will use ./config.h rather than $srcdir/config.h *
  * (which it would do because it found this file in $srcdir).           */
 #include <config.h>
+
+/*
+ * Our longest integer type:  will be a 64 bit either if long already is,
+ * or if we found some alternative such as long long.
+ * Currently we only define this to be longer than a long if --enable-lfs
+ * was given.  That enables internal use of 64-bit types even if
+ * no actual large file support is present.
+ *
+ * This comes before system.h, where we use SIZEOF_ZLONG for
+ * defining the digit buffer size.
+ */
+#ifdef ZSH_64_BIT_TYPE
+# define SIZEOF_ZLONG 8
+typedef ZSH_64_BIT_TYPE zlong;
+#ifdef ZSH_64_BIT_UTYPE
+typedef ZSH_64_BIT_UTYPE zulong;
+#else
+typedef unsigned zlong zulong;
+#endif
+#else
+# define SIZEOF_ZLONG SIZEOF_LONG
+typedef long zlong;
+typedef unsigned long zulong;
+#endif
+
 #include <system.h>
 
 /* A few typical macros */
@@ -207,6 +230,7 @@ enum {
 #define INP_HIST      (1<<2)	/* expanding history                       */
 #define INP_CONT      (1<<3)	/* continue onto previously stacked input  */
 #define INP_ALCONT    (1<<4)	/* stack is continued from alias expn.     */
+#define INP_LINENO    (1<<5)    /* update line number                      */
 
 /* Flags for metafy */
 #define META_REALLOC	0
@@ -573,9 +597,12 @@ struct job {
 #define STAT_INUSE	(1<<6)	/* this job entry is in use             */
 #define STAT_SUPERJOB	(1<<7)	/* job has a subjob                     */
 #define STAT_SUBJOB	(1<<8)	/* job is a subjob                      */
-#define STAT_CURSH	(1<<9)	/* last command is in current shell     */
-#define STAT_NOSTTY	(1<<10)	/* the tty settings are not inherited   */
+#define STAT_WASSUPER   (1<<9)  /* was a super-job, sub-job needs to be */
+				/* deleted */
+#define STAT_CURSH	(1<<10)	/* last command is in current shell     */
+#define STAT_NOSTTY	(1<<11)	/* the tty settings are not inherited   */
 				/* from this job when it exits.         */
+#define STAT_ATTACH	(1<<12)	/* delay reattaching shell to tty       */
 
 #define SP_RUNNING -1		/* fake status for jobs currently running */
 
@@ -784,20 +811,20 @@ struct param {
     union {
 	char **arr;		/* value if declared array   (PM_ARRAY)   */
 	char *str;		/* value if declared string  (PM_SCALAR)  */
-	long val;		/* value if declared integer (PM_INTEGER) */
+	zlong val;		/* value if declared integer (PM_INTEGER) */
     } u;
 
     /* pointer to function to set value of this parameter */
     union {
 	void (*cfn) _((Param, char *));
-	void (*ifn) _((Param, long));
+	void (*ifn) _((Param, zlong));
 	void (*afn) _((Param, char **));
     } sets;
 
     /* pointer to function to get value of this parameter */
     union {
 	char *(*cfn) _((Param));
-	long (*ifn) _((Param));
+	zlong (*ifn) _((Param));
 	char **(*afn) _((Param));
     } gets;
 
@@ -902,7 +929,8 @@ struct histent {
     int flags;			/* Misc flags                       */
 };
 
-#define HIST_OLD	0x00000001	/* Command is already written to disk*/
+#define HIST_OLD	0x00000002	/* Command is already written to disk*/
+#define HIST_READ	0x00000004	/* Command was read back from disk*/
 
 /* Parts of the code where history expansion is disabled *
  * should be within a pair of STOPHIST ... ALLOWHIST     */
@@ -913,7 +941,7 @@ struct histent {
 #define HISTFLAG_DONE   1
 #define HISTFLAG_NOEXEC 2
 #define HISTFLAG_RECALL 4
-
+#define HISTFLAG_SETTY  8
 
 /******************************************/
 /* Definitions for programable completion */
@@ -1108,6 +1136,7 @@ enum {
     HISTIGNOREDUPS,
     HISTIGNORESPACE,
     HISTNOSTORE,
+    HISTREDUCEBLANKS,
     HISTVERIFY,
     HUP,
     IGNOREBRACES,
@@ -1135,6 +1164,7 @@ enum {
     OVERSTRIKE,
     PATHDIRS,
     POSIXBUILTINS,
+    PRINTEIGHTBIT,
     PRINTEXITVALUE,
     PRIVILEGED,
     PROMPTCR,
@@ -1247,7 +1277,8 @@ struct ttyinfo {
 #define TCALLATTRSOFF  21
 #define TCSTANDOUTEND  22
 #define TCUNDERLINEEND 23
-#define TC_COUNT       24
+#define TCHORIZPOS     24
+#define TC_COUNT       25
 
 #define tccan(X) (tclen[X])
 
@@ -1273,14 +1304,19 @@ struct ttyinfo {
 
 #include "ztype.h"
 
-#define cmdpush(X) if (!(cmdsp >= 0 && cmdsp < 256)) {;} else cmdstack[cmdsp++]=(X)
+#define cmdpush(X) do { \
+		       if (cmdsp >= 0 && cmdsp < 256) \
+			   cmdstack[cmdsp++] = (X); \
+		   } while (0)
 #ifdef DEBUG
-# define cmdpop()  if (cmdsp <= 0) { \
-			fputs("BUG: cmdstack empty\n", stderr); \
-			fflush(stderr); \
-		   } else cmdsp--
+# define cmdpop()  do { \
+		       if (cmdsp <= 0) { \
+			   fputs("BUG: cmdstack empty\n", stderr); \
+			   fflush(stderr); \
+		       } else cmdsp--; \
+		   } while (0)
 #else
-# define cmdpop()   if (cmdsp <= 0) {;} else cmdsp--
+# define cmdpop()  do { if (cmdsp > 0) cmdsp--; } while (0)
 #endif
 
 #define CS_FOR          0
@@ -1330,8 +1366,8 @@ struct ttyinfo {
 		} while(0)
 
 # define LASTALLOC_RETURN \
-			if ((nonlocal_useheap ? global_heapalloc() : \
-			     global_permalloc()), 0) {;} else return
+			for (nonlocal_useheap ? global_heapalloc() : \
+			     global_permalloc(); 1;) return
 #else
 # define HEAPALLOC	do { int nonlocal_useheap = global_heapalloc(); \
 			alloc_stackp++; do
@@ -1345,8 +1381,8 @@ struct ttyinfo {
 		} while(0)
 
 # define LASTALLOC_RETURN \
-			if ((nonlocal_useheap ? global_heapalloc() : \
-			    global_permalloc()),alloc_stackp--,0){;}else return
+			for (nonlocal_useheap ? global_heapalloc() : \
+			     global_permalloc(); alloc_stackp--;) return
 #endif
 
 /****************/
@@ -1354,10 +1390,11 @@ struct ttyinfo {
 /****************/
 
 #ifdef DEBUG
-# define DPUTS(X,Y) if (!(X)) {;} else dputs(Y)
-# define MUSTUSEHEAP(X) if (useheap) {;} else \
-		fprintf(stderr, "BUG: permanent allocation in %s\n", X), \
-		fflush(stderr)
+# define DPUTS(X,Y) do { if ((X)) dputs(Y); } while (0)
+# define MUSTUSEHEAP(X) do { if (useheap == 0) { \
+		    fprintf(stderr, "BUG: permanent allocation in %s\n", X); \
+		    fflush(stderr); \
+		} } while (0)
 #else
 # define DPUTS(X,Y)
 # define MUSTUSEHEAP(X)
